from django.contrib import admin

from wbcommission.models import (
    Commission,
    CommissionExclusionRule,
    CommissionRole,
    CommissionRule,
    CommissionType,
)


@admin.register(CommissionType)
class CommissionTypeModelAdmin(admin.ModelAdmin):
    pass


class CommissionRuleTabularInline(admin.TabularInline):
    ordering = ("timespan__startswith", "assets_under_management_range__startswith")
    fields = (
        "timespan",
        "assets_under_management_range",
        "percent",
    )
    model = CommissionRule
    extra = 0


class CommissionRoleTabularInline(admin.TabularInline):
    model = CommissionRole
    autocomplete_fields = ["person"]
    fields = (
        "person",
        "commission",
    )
    extra = 0


class CommissionTabularInline(admin.TabularInline):
    extra = 0
    model = Commission
    fields = [
        "order",
        "crm_recipient",
        "account_role_type_recipient",
        "portfolio_role_recipient",
        "commission_type",
        "net_commission",
        "is_hidden",
        "exclusion_rule_account_role_type",
    ]
    readonly_fields = ["order"]
    ordering = ["order"]
    autocomplete_fields = ["account", "crm_recipient"]
    show_change_link = True
    fk_name = "account"


@admin.register(Commission)
class CommissionModelAdmin(admin.ModelAdmin):
    list_display = [
        "account",
        "crm_recipient",
        "portfolio_role_recipient",
        "account_role_type_recipient",
        "order",
        "net_commission",
        "commission_type",
        "is_hidden",
        "exclusion_rule_account_role_type",
    ]

    autocomplete_fields = ["account", "crm_recipient"]

    inlines = [CommissionRoleTabularInline, CommissionRuleTabularInline]


@admin.register(CommissionExclusionRule)
class CommissionExclusionRuleAdmin(admin.ModelAdmin):
    list_display = [
        "product",
        "commission_type",
        "account_role_type",
        "timespan",
        "overriding_percent",
        "overriding_net_or_gross_commission",
    ]
    autocomplete_fields = ["product", "account_role_type"]



from datetime import date, datetime

from django.contrib.messages import warning
from django.http import HttpRequest
from django.utils.functional import cached_property

from wbcommission.models import Commission, Rebate


class CommissionPermissionMixin:
    queryset = Commission.objects.all()
    request: HttpRequest

    @cached_property
    def validity_date(self) -> date:
        if validity_date_repr := self.request.GET.get("validity_date"):
            return datetime.strptime(validity_date_repr, "%Y-%m-%d")
        return date.today()

    def get_queryset(self):
        # allow the user to see claim only on accounts it can see
        queryset = super().get_queryset()
        base_commission_count = queryset.count()

        filtered_queryset = super().get_queryset().filter_for_user(self.request.user, validity_date=self.validity_date)
        if base_commission_count != filtered_queryset.count():
            warning(
                self.request,
                "Certain commission rules are not visible due to insufficient permissions. Contact your administrator if you believe you should have access.",
            )
        return filtered_queryset


class RebatePermissionMixin:
    queryset = Rebate.objects.all()
    request: HttpRequest

    @cached_property
    def validity_date(self) -> date:
        if validity_date_repr := self.request.GET.get("validity_date"):
            return datetime.strptime(validity_date_repr, "%Y-%m-%d")
        return date.today()

    def get_queryset(self):
        # allow the user to see claim only on accounts it can see
        return (
            super().get_queryset().filter_for_user(self.request.user, validity_date=self.validity_date)  # type: ignore
        )



from wbcore import serializers as wb_serializers

from wbcommission.models import Rebate


class RebateModelSerializer(wb_serializers.ModelSerializer):
    id = wb_serializers.PrimaryKeyField()
    value_usd = wb_serializers.FloatField(read_only=True)

    class Meta:
        model = Rebate
        fields = ("id", "value_usd")


class RebateProductMarginalitySerializer(wb_serializers.Serializer):
    id = wb_serializers.PrimaryKeyField()
    title = wb_serializers.CharField(read_only=True)
    currency_symbol = wb_serializers.CharField(read_only=True)
    sum_management_fees = wb_serializers.DecimalField(max_digits=16, decimal_places=2, label="Management Fees")
    sum_performance_fees = wb_serializers.DecimalField(max_digits=16, decimal_places=2, label="Management Rebates")
    sum_crystalized_performance_fees = wb_serializers.DecimalField(
        max_digits=16, decimal_places=2, label="Crystalized Performance Fees"
    )
    sum_management_rebates = wb_serializers.DecimalField(max_digits=16, decimal_places=2, label="Performance Fees")
    sum_performance_rebates = wb_serializers.DecimalField(max_digits=16, decimal_places=2, label="Performance Rebates")
    marginality_management = wb_serializers.DecimalField(
        max_digits=16, decimal_places=2, label="Marginality Management"
    )
    marginality_performance = wb_serializers.DecimalField(
        max_digits=16, decimal_places=2, label="Marginality Performance"
    )

    total_fees = wb_serializers.DecimalField(max_digits=16, decimal_places=2, label="Total Fees")
    total_rebates = wb_serializers.DecimalField(max_digits=16, decimal_places=2, label="Total Rebates")
    total_marginality = wb_serializers.DecimalField(max_digits=16, decimal_places=2, label="Total Marginality")

    total_fees_usd = wb_serializers.DecimalField(max_digits=16, decimal_places=2, label="Total Fees (USD)")
    total_rebates_usd = wb_serializers.DecimalField(max_digits=16, decimal_places=2, label="Total Rebates (USD)")
    total_usd = wb_serializers.DecimalField(max_digits=16, decimal_places=2, label="Total Net (USD)")

    class Meta:
        percent_fields = [
            "marginality_management",
            "marginality_performance",
            "total_marginality",
        ]
        decorators = {
            "sum_management_fees": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{currency_symbol}}"
            ),
            "sum_performance_fees": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{currency_symbol}}"
            ),
            "sum_management_rebates": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{currency_symbol}}"
            ),
            "sum_performance_rebates": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{currency_symbol}}"
            ),
            "total_fees": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{currency_symbol}}"
            ),
            "total_rebates": wb_serializers.decorator(
                decorator_type="text", position="left", value="{{currency_symbol}}"
            ),
            "total_fees_usd": wb_serializers.decorator(decorator_type="text", position="left", value="$"),
            "total_rebates_usd": wb_serializers.decorator(decorator_type="text", position="left", value="$"),
        }

        fields = (
            "id",
            "title",
            "currency_symbol",
            "sum_management_fees",
            "sum_performance_fees",
            "sum_crystalized_performance_fees",
            "sum_management_rebates",
            "sum_performance_rebates",
            "marginality_management",
            "marginality_performance",
            "total_fees",
            "total_rebates",
            "total_marginality",
            "total_fees_usd",
            "total_rebates_usd",
            "total_usd",
        )
        read_only_fields = fields



# Generated by Django 4.1.9 on 2023-07-05 14:26

from decimal import Decimal

import django.contrib.postgres.constraints
import django.contrib.postgres.fields.ranges
import django.db.models.deletion
import wbcommission.models.rebate
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcrm", "0005_account_accountrole_accountroletype_and_more"),
        ("wbcommission", "0001_initial"),
        ("directory", "0004_entry_is_draft_entry"),
    ]

    operations = [
        migrations.RenameModel(old_name="Commission", new_name="OldCommission"),
        migrations.DeleteModel(name="rebate"),
        migrations.CreateModel(
            name="CommissionType",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=256, verbose_name="Name")),
                ("key", models.CharField(max_length=256, unique=True)),
            ],
        ),
        migrations.CreateModel(
            name="Commission",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "portfolio_role_recipient",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("MANAGER", "Manager"),
                            ("RISK_MANAGER", "Risk Manager"),
                            ("PORTFOLIO_MANAGER", "Portfolio Manager"),
                            ("ANALYST", "Analyst"),
                        ],
                        max_length=32,
                        null=True,
                        verbose_name="Recipient Role",
                    ),
                ),
                ("order", models.IntegerField()),
                ("net_commission", models.BooleanField(default=True, verbose_name="Net Commission Rule")),
                (
                    "commission_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="commissions",
                        to="wbcommission.commissiontype",
                        verbose_name="Commission Type",
                    ),
                ),
                (
                    "is_hidden",
                    models.BooleanField(
                        default=True,
                        help_text="If False, this commission rule can be seen by anoyone that can access the related account. Otherwise, only an explicit role will grant access",
                        verbose_name="Public Commission Rule",
                    ),
                ),
                (
                    "account",
                    models.ForeignKey(
                        limit_choices_to=models.Q(("is_terminal_account", True)),
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="account_commissions",
                        to="wbcrm.account",
                        verbose_name="Account",
                    ),
                ),
                (
                    "account_role_type_recipient",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="recipient_commissions",
                        to="wbcrm.accountroletype",
                        verbose_name="Account Role Type Recipient",
                    ),
                ),
                (
                    "crm_recipient",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="recipient_commissions",
                        to="directory.entry",
                        verbose_name="Recipient",
                    ),
                ),
            ],
            options={
                "verbose_name": "Commission",
                "verbose_name_plural": "Commissions",
            },
        ),
        migrations.AddConstraint(
            model_name="commission",
            constraint=models.UniqueConstraint(
                fields=("commission_type", "account", "crm_recipient"), name="unique_crm_recipient_account"
            ),
        ),
        migrations.AddConstraint(
            model_name="commission",
            constraint=models.UniqueConstraint(
                fields=("commission_type", "account", "portfolio_role_recipient"),
                name="unique_portfolio_role_recipient_account",
            ),
        ),
        migrations.AddConstraint(
            model_name="commission",
            constraint=models.UniqueConstraint(
                fields=("commission_type", "account", "account_role_type_recipient"),
                name="unique_account_role_type_recipient_account",
            ),
        ),
        migrations.AddConstraint(
            model_name="commission",
            constraint=models.CheckConstraint(
                check=models.Q(
                    models.Q(
                        ("account_role_type_recipient__isnull", True),
                        ("crm_recipient__isnull", False),
                        ("portfolio_role_recipient__isnull", True),
                    ),
                    models.Q(
                        ("account_role_type_recipient__isnull", True),
                        ("crm_recipient__isnull", True),
                        ("portfolio_role_recipient__isnull", False),
                    ),
                    models.Q(
                        ("account_role_type_recipient__isnull", False),
                        ("crm_recipient__isnull", True),
                        ("portfolio_role_recipient__isnull", True),
                    ),
                    _connector="OR",
                ),
                name="Only one recipient type set",
            ),
        ),
        migrations.CreateModel(
            name="CommissionRole",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("computed_str", models.CharField(blank=True, max_length=512, null=True, verbose_name="Name")),
                (
                    "commission",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="roles",
                        to="wbcommission.commission",
                        verbose_name="Commission Line",
                    ),
                ),
                (
                    "person",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="rebates_account_roles",
                        to="directory.person",
                        verbose_name="Person",
                    ),
                ),
            ],
            options={
                "verbose_name": "Commission Role",
                "verbose_name_plural": "Commission Roles",
            },
        ),
        migrations.CreateModel(
            name="CommissionRule",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("computed_str", models.CharField(blank=True, max_length=512, null=True, verbose_name="Name")),
                ("timespan", django.contrib.postgres.fields.ranges.DateRangeField(verbose_name="Timespan")),
                (
                    "assets_under_management_range",
                    django.contrib.postgres.fields.ranges.DecimalRangeField(verbose_name="AUM Range"),
                ),
                ("percent", models.FloatField(default=0, verbose_name="Percent")),
            ],
        ),
        migrations.AddField(
            model_name="commissionrule",
            name="commission",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="rules",
                to="wbcommission.commission",
                verbose_name="Commission Line",
            ),
        ),
        migrations.AddConstraint(
            model_name="commissionrule",
            constraint=django.contrib.postgres.constraints.ExclusionConstraint(
                expressions=[("timespan", "&&"), ("assets_under_management_range", "&&"), ("commission", "=")],
                name="exclude_overlapping_rules",
            ),
        ),
        migrations.CreateModel(
            name="CommissionExclusionRule",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("timespan", django.contrib.postgres.fields.ranges.DateRangeField(verbose_name="Timespan")),
                (
                    "overriding_percent",
                    models.DecimalField(decimal_places=3, max_digits=4, verbose_name="Overriding Percent"),
                ),
                (
                    "overriding_net_or_gross_commission",
                    models.CharField(
                        choices=[("NET", "Net"), ("GROSS", "Gross"), ("DEFAULT", "Default")],
                        default="DEFAULT",
                        max_length=16,
                        verbose_name="Overriding Net or Gross Commission Rule",
                    ),
                ),
                (
                    "account_role_type",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="wbcrm.accountroletype",
                        verbose_name="Account Role Type Recipient",
                    ),
                ),
                (
                    "commission_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="commission_exclusion_rules",
                        to="wbcommission.commissiontype",
                        verbose_name="Commission Type",
                    ),
                ),
                (
                    "product",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="wbportfolio.product", verbose_name="Product"
                    ),
                ),
            ],
            options={
                "verbose_name": "Commission Exclusion Rule",
                "verbose_name_plural": "Commissions Exclusion Rules",
            },
        ),
        migrations.AddConstraint(
            model_name="commissionexclusionrule",
            constraint=django.contrib.postgres.constraints.ExclusionConstraint(
                expressions=[
                    ("timespan", "&&"),
                    ("product", "="),
                    ("commission_type", "="),
                    ("account_role_type", "="),
                ],
                name="exclude_overlapping_exclusion_rules",
            ),
        ),
        migrations.AddField(
            model_name="commission",
            name="exclusion_rule_account_role_type",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="wbcrm.accountroletype",
                verbose_name="Account Role to decide with exclusion rule applies",
            ),
        ),
        migrations.CreateModel(
            name="Rebate",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("date", models.DateField(verbose_name="Date")),
                (
                    "value",
                    models.DecimalField(decimal_places=4, default=Decimal(0.0), max_digits=16, verbose_name="Value"),
                ),
                (
                    "account",
                    models.ForeignKey(
                        limit_choices_to=models.Q(("is_terminal_account", True)),
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="rebates",
                        to="wbcrm.account",
                        verbose_name="Account",
                    ),
                ),
                (
                    "commission",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="rebates",
                        to="wbcommission.commission",
                        verbose_name="Commission Line",
                    ),
                ),
                (
                    "commission_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="rebates",
                        to="wbcommission.commissiontype",
                        verbose_name="Commission Type",
                    ),
                ),
                (
                    "product",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="rebates",
                        to="wbportfolio.product",
                        verbose_name="Product",
                    ),
                ),
                (
                    "recipient",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="recipient_rebates",
                        to="directory.entry",
                        verbose_name="Recipient",
                    ),
                ),
            ],
            options={
                "verbose_name": "Rebate",
                "verbose_name_plural": "Rebates",
                "unique_together": {("date", "recipient", "account", "product", "commission_type")},
                "index_together": {("commission_type", "date", "recipient", "product", "account")},
            },
            bases=(wbcommission.models.rebate.BookingEntryCalculatedValueMixin, models.Model),
        ),
        migrations.RemoveField(
            model_name="accountcustomer",
            name="account",
        ),
        migrations.RemoveField(
            model_name="accountcustomer",
            name="entry",
        ),
        migrations.RemoveField(
            model_name="accountcustomercommissionconstraint",
            name="customer",
        ),
        migrations.RemoveField(
            model_name="accountcustomercommissionconstraint",
            name="products",
        ),
        migrations.DeleteModel(
            name="AccountCustomer",
        ),
        migrations.DeleteModel(
            name="AccountCustomerCommissionConstraint",
        ),
        migrations.DeleteModel(
            name="OldCommission",
        ),
        migrations.DeleteModel(
            name="CommissionAccountRole",
        ),
        migrations.AlterField(
            model_name="commission",
            name="is_hidden",
            field=models.BooleanField(
                default=True,
                help_text="If False, this commission rule can be seen by anoyone that can access the related account. Otherwise, only an explicit role will grant access",
                verbose_name="Hidden Commission Rule",
            ),
        ),
        migrations.AlterField(
            model_name="commissionrule",
            name="percent",
            field=models.DecimalField(decimal_places=3, default=Decimal("0"), max_digits=4, verbose_name="Percent"),
        ),
        migrations.AlterModelOptions(
            name="commission",
            options={
                "permissions": (("administrate_commission", "Can administrate Commission"),),
                "verbose_name": "Commission",
                "verbose_name_plural": "Commissions",
            },
        ),
    ]



# Generated by Django 5.0.9 on 2024-10-17 07:44

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbcommission", "0005_alter_rebate_audit_log"),
    ]

    operations = [
        migrations.AddField(
            model_name="commissionrule",
            name="consider_zero_percent_for_exclusion",
            field=models.BooleanField(
                default=False,
                help_text="If true, the commission rule with percent 0 (initially consider disabled), will be considered for exclusion rule matching",
                verbose_name="Consider 0 percent for exclusion",
            ),
        ),
    ]



from .rebate import (
    CustomerRebateGroupByFilter,
    RebateDateFilter,
    RebateGroupByFilter,
    RebateMarginalityFilter,
)
from .signals import *
from .commissions import CommissionAccountFilterSet, CommissionRuleCommissionFilterSet



from typing import Optional

from django.dispatch import receiver
from rest_framework.reverse import reverse
from wbcore.contrib.color.enums import WBColor
from wbcore.enums import Unit
from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display import Page, Style, add_display_pages
from wbcore.metadata.configs.display.instance_display.layouts.layouts import Layout
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    Display,
    create_simple_display,
    create_simple_section,
    default,
)
from wbcore.metadata.configs.display.view_config import DisplayViewConfig
from wbcrm.viewsets.display import AccountDisplayConfig

from wbcommission.models import CommissionType


class CommissionTypeDisplayConfigClass(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="name", label="Name"),
                dp.Field(key="key", label="Key"),
            ]
        )

    def get_instance_display(self) -> Display:
        return create_simple_display([["name", "key"]])


class CommissionDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            legends=[
                dp.Legend(
                    items=[
                        dp.LegendItem(
                            icon=WBColor.BLUE.value,
                            label="Simulated percent",
                        ),
                    ],
                ),
            ],
            fields=[
                dp.Field(key="order", label="Order", width=Unit.PIXEL(75)),
                dp.Field(
                    key=None,
                    label="Recipient",
                    open_by_default=False,
                    children=[
                        dp.Field(key="recipient_repr", label="Group", width=Unit.PIXEL(250)),
                        dp.Field(
                            key="recipient_details", label="Simulated Details", width=Unit.PIXEL(250), show="open"
                        ),
                    ],
                ),
                dp.Field(
                    key=None,
                    label="Rules",
                    open_by_default=True,
                    children=[
                        dp.Field(
                            key="percent",
                            label="Percent",
                            width=Unit.PIXEL(100),
                            formatting_rules=[
                                dp.FormattingRule(
                                    condition=[("∃", True, "recipient_repr")],
                                    style={"color": WBColor.BLUE.value, "fontWeight": "bold"},
                                )
                            ],
                        ),
                        dp.Field(key="net_commission", label="Net", width=Unit.PIXEL(100)),
                        dp.Field(key="timespan", label="Timespan", show="open", width=Unit.PIXEL(300)),
                        dp.Field(
                            key="assets_under_management_range", label="AUM range", show="open", width=Unit.PIXEL(300)
                        ),
                    ],
                ),
                dp.Field(
                    key=None,
                    label="Permission",
                    open_by_default=False,
                    children=[
                        dp.Field(key="is_hidden", label="Hidden", width=Unit.PIXEL(100)),
                        dp.Field(key="roles_repr", label="Direct Roles", width=Unit.PIXEL(350), show="open"),
                    ],
                ),
                dp.Field(key="exclusion_rule_account_role_type", label="Exclusion rule type"),
            ],
            tree=True,
            tree_group_field="order",
            tree_group_field_sortable=False,
            tree_group_level_options=[
                dp.TreeGroupLevelOption(
                    filter_key="commission",
                    filter_depth=1,
                    list_endpoint=reverse(
                        "wbcommission:commissionrule-list",
                        args=[],
                        request=self.request,
                    ),
                )
            ],
        )

    def get_instance_display(self) -> Display | None:
        return create_simple_display(
            [
                ["crm_recipient", "portfolio_role_recipient", "account_role_type_recipient"],
                ["net_commission", "is_hidden", "exclusion_rule_account_role_type"],
                ["rule_section", "rule_section", "rule_section"],
            ],
            sections=[
                create_simple_section(
                    "rule_section",
                    "Validity",
                    [["rules"]],
                    "rules",
                    collapsed=False,
                )
            ],
        )


class CommissionRuleCommissionDisplayConfig(DisplayViewConfig):
    def get_list_display(self) -> Optional[dp.ListDisplay]:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="percent", label="Percent"),
                dp.Field(key="net_commission", label="Net"),
                dp.Field(key="timespan", label="Timespan"),
                dp.Field(key="assets_under_management_range", label="AUM range"),
            ]
        )

    def get_instance_display(self) -> Display | None:
        return create_simple_display(
            [
                ["percent", "consider_zero_percent_for_exclusion"],
                ["timespan", "assets_under_management_range"],
            ]
        )


@receiver(add_display_pages, sender=AccountDisplayConfig)
def add_commission_page(sender, *args, **kwargs):
    grid_template_areas = []
    sections = []
    grid_template_rows = []
    ct_count = CommissionType.objects.count()
    for ct in CommissionType.objects.all():
        key = "commission-" + ct.key
        grid_template_areas.append([key])
        sections.append(
            create_simple_section(
                key=key,
                title=ct.name,
                grid_template_areas=[[key]],
                inline_key=key,
                collapsible=False,
            )
        )
        grid_template_rows.append(Style.pct(int(100 / ct_count)))
    return [
        Page(
            title="Commission",
            layouts={
                default(): Layout(
                    grid_auto_rows=Style.AUTO,
                    grid_template_areas=grid_template_areas,
                    grid_template_rows=grid_template_rows,
                    sections=sections,
                )
            },
        ),
    ]



from django.core.exceptions import ValidationError
from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.directory.models import Entry
from wbcore.contrib.directory.serializers import EntryRepresentationSerializer
from wbcore.signals import add_additional_resource
from wbcrm.models import AccountRoleType
from wbcrm.serializers import (
    AccountModelSerializer,
    AccountRepresentationSerializer,
    AccountRoleTypeRepresentationSerializer,
)
from wbportfolio.models import PortfolioRole

from wbcommission.models import Commission, CommissionRule, CommissionType


class CommissionRepresentationSerializer(wb_serializers.RepresentationSerializer):
    class Meta:
        model = Commission
        fields = ("id",)


class CommissionTypeRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbcommission:commissiontype-detail")

    class Meta:
        model = CommissionType
        fields = (
            "id",
            "name",
            "key",
            "_detail",
        )


class CommissionTypeModelSerializer(wb_serializers.ModelSerializer):
    class Meta:
        model = CommissionType
        read_only_fields = ("key",)
        fields = (
            "id",
            "name",
            "key",
        )


class CommissionModelSerializer(wb_serializers.ModelSerializer):
    _group_key = wb_serializers.CharField(read_only=True, label="Group Key")
    _account = AccountRepresentationSerializer(source="account")
    crm_recipient = wb_serializers.PrimaryKeyRelatedField(
        queryset=Entry.objects.all(), read_only=lambda view: not view.new_mode, required=False
    )
    _crm_recipient = EntryRepresentationSerializer(source="crm_recipient", read_only=lambda view: not view.new_mode)
    account_role_type_recipient = wb_serializers.PrimaryKeyRelatedField(
        queryset=AccountRoleType.objects.all(), read_only=lambda view: not view.new_mode, required=False
    )
    _account_role_type_recipient = AccountRoleTypeRepresentationSerializer(
        source="account_role_type_recipient", read_only=lambda view: not view.new_mode
    )
    portfolio_role_recipient = wb_serializers.ChoiceField(
        choices=PortfolioRole.RoleType.choices, read_only=lambda view: not view.new_mode, required=False
    )

    _commission_type = CommissionTypeRepresentationSerializer(source="commission_type")
    _exclusion_rule_account_role_type = AccountRoleTypeRepresentationSerializer(
        source="exclusion_rule_account_role_type"
    )

    roles_repr = wb_serializers.CharField(
        read_only=True,
        label="Direct Roles",
        help_text="The direct role this commission rule plays in defining rebate and commission accessibility.",
    )
    percent = wb_serializers.SerializerMethodField(
        read_only=True,
        field_class=wb_serializers.FloatField,
        percent=True,
        help_text="The raw rule percentage. If a simulation product and date are selected, it displays the simulated percentage (in blue).",
    )
    recipient_repr = wb_serializers.SerializerMethodField(
        read_only=True, help_text="The raw recipient group to which this rule applies."
    )
    recipient_details = wb_serializers.SerializerMethodField(
        read_only=True,
        help_text="Preview recipients and their allocated percentages based on the selected product and date (if specified).",
    )
    timespan = wb_serializers.DateRangeField(
        read_only=True, label="Timespan", help_text="The date range at which the rule applies."
    )
    assets_under_management_range = wb_serializers.DecimalRangeField(
        read_only=True, label="AUM Range", help_text="The AUM range at which the rule applies."
    )

    def validate(self, data):
        crm_recipient = data.get("crm_recipient")
        account_role_type_recipient = data.get("account_role_type_recipient")
        portfolio_role_recipient = data.get("portfolio_role_recipient")
        if not crm_recipient and not account_role_type_recipient and not portfolio_role_recipient:
            raise ValidationError(
                {
                    ("crm_recipient"): _("Exactly one recipient field must be set (others left empty)."),
                    ("account_role_type_recipient"): _("Exactly one recipient field must be set (others left empty)."),
                    ("portfolio_role_recipient"): _("Exactly one recipient field must be set (others left empty)."),
                }
            )
        if not self.instance and (view := self.context.get("view")):
            if (
                crm_recipient
                and Commission.objects.filter(
                    account=view.account, commission_type=view.commission_type, crm_recipient=crm_recipient
                ).exists()
            ):
                raise ValidationError({"crm_recipient": "A commission line already exists for this crm recipient."})
            if (
                account_role_type_recipient
                and Commission.objects.filter(
                    account=view.account,
                    commission_type=view.commission_type,
                    account_role_type_recipient=account_role_type_recipient,
                ).exists()
            ):
                raise ValidationError(
                    {"account_role_type_recipient": "A commission line already exists for this account role type."}
                )
            if (
                portfolio_role_recipient
                and Commission.objects.filter(
                    account=view.account,
                    commission_type=view.commission_type,
                    portfolio_role_recipient=portfolio_role_recipient,
                ).exists()
            ):
                raise ValidationError(
                    {"portfolio_role_recipient": "A commission line already exists for this portfolio role type."}
                )
        return data

    def get_recipient_repr(self, obj):
        return obj.get_recipient_repr()

    def get_recipient_details(self, obj):
        res = ""
        if (view := self.context.get("view")) and view.simulation_date and view.simulation_product:
            recipients = obj.get_recipients(view.account, view.simulation_product, view.simulation_date)
            res = ", ".join(map(lambda o: f"{str(o[0])} ({o[1]:.2%})", recipients))
        return res

    def get_percent(self, obj):
        if (view := self.context.get("view")) and (actual_percent := view.actual_percents) is not None:
            return actual_percent.get(obj, 0)
        return None

    def get_unique_together_validators(self):
        return []

    @wb_serializers.register_resource()
    def additional_resources(self, instance, request, user):
        return {
            "rules": reverse("wbcommission:commission-rule-list", args=[instance.id]),
        }

    class Meta:
        model = Commission

        fields = (
            "id",
            "_group_key",
            "order",
            "account",
            "_account",
            "crm_recipient",
            "_crm_recipient",
            "portfolio_role_recipient",
            "account_role_type_recipient",
            "_account_role_type_recipient",
            "_commission_type",
            "net_commission",
            "commission_type",
            "is_hidden",
            "exclusion_rule_account_role_type",
            "_exclusion_rule_account_role_type",
            "roles_repr",
            "recipient_repr",
            "recipient_details",
            "percent",
            "timespan",
            "assets_under_management_range",
            "_additional_resources",
        )


class CommissionRuleCommissionModelSerializer(wb_serializers.ModelSerializer):
    net_commission = wb_serializers.BooleanField(read_only=True, label="Net Commission")

    class Meta:
        model = CommissionRule
        fields = (
            "id",
            "timespan",
            "net_commission",
            "assets_under_management_range",
            "percent",
            "consider_zero_percent_for_exclusion",
            "commission",
        )


@receiver(add_additional_resource, sender=AccountModelSerializer)
def add_commission_resource(sender, serializer, instance, request, user, **kwargs):
    res = {}
    for ct in CommissionType.objects.all():
        res[f"commission-{ct.key}"] = reverse("wbcommission:type-account-commission-list", args=[ct.id, instance.id])
    return res



from .rebate import RebateModelSerializer, RebateProductMarginalitySerializer
from .commissions import CommissionTypeRepresentationSerializer, CommissionTypeModelSerializer, CommissionModelSerializer, CommissionRepresentationSerializer
from .signals import *



from datetime import date

from django.contrib.messages import info
from django.contrib.postgres.aggregates import StringAgg
from django.db.models import F
from django.shortcuts import get_object_or_404
from django.utils.dateparse import parse_date
from django.utils.functional import cached_property
from rest_framework.exceptions import AuthenticationFailed
from wbcore import viewsets
from wbcore.viewsets.mixins import OrderableMixin
from wbcrm.models import Account
from wbportfolio.models import Product

from wbcommission.models import Commission, CommissionRule, CommissionType

from ..filters import CommissionAccountFilterSet, CommissionRuleCommissionFilterSet
from ..models.account_service import AccountRebateManager
from ..serializers import (
    CommissionModelSerializer,
    CommissionTypeModelSerializer,
    CommissionTypeRepresentationSerializer,
)
from ..serializers.commissions import CommissionRepresentationSerializer, CommissionRuleCommissionModelSerializer
from .buttons import CommissionButtonConfig
from .display import CommissionDisplayConfig, CommissionRuleCommissionDisplayConfig, CommissionTypeDisplayConfigClass
from .endpoints import CommissionEndpointConfig, CommissionRuleCommissionEndpointConfig
from .mixins import CommissionPermissionMixin
from .titles import CommissionTitleConfig


class CommissionTypeRepresentationModelViewSet(viewsets.RepresentationViewSet):
    serializer_class = CommissionTypeRepresentationSerializer
    queryset = CommissionType.objects.all()
    ordering_fields = search_fields = ordering = ["name"]


class CommissionTypeModelViewSet(viewsets.ModelViewSet):
    serializer_class = CommissionTypeModelSerializer
    queryset = CommissionType.objects.all()
    ordering_fields = search_fields = ordering = ["name"]
    display_config_class = CommissionTypeDisplayConfigClass


class CommissionRepresentationModelViewSet(viewsets.RepresentationViewSet):
    serializer_class = CommissionRepresentationSerializer
    queryset = Commission.objects.all()
    ordering_fields = search_fields = ordering = ["id"]


class CommissionModelViewSet(OrderableMixin, CommissionPermissionMixin, viewsets.ModelViewSet):
    display_config_class = CommissionDisplayConfig
    title_config_class = CommissionTitleConfig
    endpoint_config_class = CommissionEndpointConfig
    button_config_class = CommissionButtonConfig

    serializer_class = CommissionModelSerializer

    filterset_class = CommissionAccountFilterSet
    queryset = Commission.objects.select_related(
        "account",
        "crm_recipient",
        "account_role_type_recipient",
        "commission_type",
    )
    ordering = ["order"]

    @cached_property
    def account(self) -> Account:
        return get_object_or_404(Account, pk=self.kwargs["account_id"])

    @cached_property
    def commission_type(self) -> CommissionType:
        return get_object_or_404(CommissionType, pk=self.kwargs["commission_type_id"])

    @cached_property
    def simulation_date(self) -> date | None:
        if compute_date_str := self.request.GET.get("only_valid_at_date"):
            return parse_date(compute_date_str)

    @cached_property
    def simulation_product(self) -> Product | None:
        if product_id := self.request.GET.get("simulate_for_product"):
            return Product.objects.get(id=product_id)

    @cached_property
    def actual_percents(self) -> dict | None:
        if self.simulation_product and self.simulation_date:
            rebate_manager = AccountRebateManager(
                self.account.get_root(), self.commission_type.key
            )  # This will be loaded dynamically in a future iteration
            rebate_manager.initialize()
            idx = rebate_manager.df_fees.index.get_level_values(1)
            latest_valid_date = idx[idx <= self.simulation_date].max()
            if latest_valid_date != self.simulation_date:
                info(
                    self.request,
                    f"To simulate the recipient percent, We had to move the date to the latest valid date ({self.simulation_date:%Y-%m-%d} -> {latest_valid_date:%Y-%m-%d})",
                )

            terminal_accounts_total_holding = rebate_manager.get_terminal_accounts_total_holding(latest_valid_date)
            terminal_account_holding_ratio = rebate_manager.get_terminal_account_holding_ratio(
                self.account,
                self.simulation_product,
                latest_valid_date,
            )
            res = {}
            if terminal_account_holding_ratio:
                res = dict(
                    self.commission_type.get_valid_commissions(
                        self.account, latest_valid_date, self.simulation_product, terminal_accounts_total_holding
                    )
                )
            return res

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(commission_type=self.commission_type, account__in=self.account.get_ancestors(include_self=True))
            .annotate(
                _group_key=F("id"),
                roles_repr=StringAgg(F("roles__person__computed_str"), delimiter=", "),
            )
        )


class CommissionRuleCommissionModelViewSet(viewsets.ModelViewSet):
    display_config_class = CommissionRuleCommissionDisplayConfig
    endpoint_config_class = CommissionRuleCommissionEndpointConfig

    serializer_class = CommissionRuleCommissionModelSerializer
    filterset_class = CommissionRuleCommissionFilterSet
    queryset = CommissionRule.objects.select_related("commission")
    ordering = ["percent", "id"]

    @cached_property
    def commission(self) -> Commission:
        commission_id = self.kwargs.get("commission_id", self.request.GET.get("commission"))
        if Commission.objects.filter(id=commission_id).filter_for_user(self.request.user).exists():
            return get_object_or_404(Commission, pk=commission_id)
        raise AuthenticationFailed()

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(commission=self.commission)
            .annotate(net_commission=F("commission__net_commission"))
        )



from wbcore.metadata.configs.titles import TitleViewConfig


class CommissionTitleConfig(TitleViewConfig):
    def get_list_title(self) -> str | None:
        return f"{self.view.commission_type} Commission Rules for account {self.view.account}"

    def get_create_title(self) -> str | None:
        return f"Create {self.view.commission_type} Commission rule for account {self.view.account}"



from django.apps import apps
from django.urls import include, path
from wbcore.routers import WBCoreRouter

from wbcommission import viewsets

router = WBCoreRouter()

router.register(
    r"rebatemarginalitytable",
    viewsets.RebateProductMarginalityViewSet,
    basename="rebatemarginalitytable",
)
router.register(
    r"rebatetable",
    viewsets.RebatePandasView,
    basename="rebatetable",
)
router.register(
    r"rebate",
    viewsets.RebateModelViewSet,
    basename="rebate",
)
router.register(
    r"commissiontype",
    viewsets.CommissionTypeModelViewSet,
    basename="commissiontype",
)
router.register(
    r"commissiontyperepresentation",
    viewsets.CommissionTypeRepresentationModelViewSet,
    basename="commissiontyperepresentation",
)
router.register(
    r"commissionrepresentation",
    viewsets.CommissionRepresentationModelViewSet,
    basename="commissionrepresentation",
)
router.register(
    r"commissionrule",
    viewsets.CommissionRuleCommissionModelViewSet,
    basename="commissionrule",
)
commission_type_account_router = WBCoreRouter()
commission_type_account_router.register(
    "commission", viewsets.CommissionModelViewSet, basename="type-account-commission"
)

commission_router = WBCoreRouter()
commission_router.register("rule", viewsets.CommissionRuleCommissionModelViewSet, basename="commission-rule")


if apps.is_installed("wbportfolio"):
    from wbportfolio.viewsets.transactions.claim import ClaimAPIModelViewSet

    router.register(r"claim-api", ClaimAPIModelViewSet, basename="claim-api")

urlpatterns = [
    path("", include(router.urls)),
    path("type/<int:commission_type_id>/account/<int:account_id>/", include(commission_type_account_router.urls)),
    path("commission/<int:commission_id>/", include(commission_router.urls)),
]



from .rebate import (
    RebatePandasViewDisplayConfig,
    RebateProductMarginalityDisplayConfig,
)
from .commissions import CommissionTypeDisplayConfigClass, CommissionDisplayConfig, CommissionRuleCommissionDisplayConfig



import math
from contextlib import suppress
from datetime import date
from decimal import Decimal
from typing import TYPE_CHECKING, Generator, Optional

from django.contrib.postgres.constraints import ExclusionConstraint
from django.contrib.postgres.fields import (
    DateRangeField,
    DecimalRangeField,
    RangeOperators,
)
from django.db import models
from django.db.models import Exists, OuterRef, Q, QuerySet, Sum, UniqueConstraint
from django.db.models.signals import post_save
from django.dispatch import receiver
from ordered_model.models import OrderedModel, OrderedModelManager, OrderedModelQuerySet
from psycopg.types.range import DateRange, NumericRange
from tqdm import tqdm
from wbcore.contrib.directory.models import Entry
from wbcore.models import WBModel
from wbcore.signals import pre_merge
from wbcore.utils.models import ComplexToStringMixin
from wbcrm.models.accounts import (
    Account,
    AccountRole,
    AccountRoleType,
    AccountRoleValidity,
)
from wbportfolio.models import PortfolioRole

from .account_service import AccountRebateManager

if TYPE_CHECKING:
    from django.db.models.manager import RelatedManager

from wbcore.contrib.authentication.models import User


class CommissionDefaultQueryset(OrderedModelQuerySet):
    def filter_for_user(self, user: User, validity_date: date | None = None) -> QuerySet:
        """
        Protect the chained queryset and filter the commission this user cannot see based on the following rules:

        * not-hidden commission: Ever user with a direct valid account role on the commission's account
        * hidden: only user with a direct commission role on that commission line
        * in any case, all user with direct commission role or recipient of the commission
        """
        if not validity_date:
            validity_date = date.today()
        if user.has_perm("wbcommission.administrate_commission"):
            return self
        valid_accounts = Account.objects.filter_for_user(user, validity_date=validity_date, strict=True)
        return self.annotate(
            can_see_account=Exists(valid_accounts.filter(id=OuterRef("account"))),
            has_direct_commission_role=Exists(
                CommissionRole.objects.filter(person=user.profile, commission=OuterRef("pk"))
            ),
        ).filter(
            (Q(is_hidden=False) & Q(can_see_account=True))
            | Q(has_direct_commission_role=True)
            | Q(crm_recipient=user.profile.entry_ptr)
        )


class CommissionManager(OrderedModelManager):
    def get_queryset(self) -> CommissionDefaultQueryset:
        return CommissionDefaultQueryset(self.model)

    def filter_for_user(self, user: User, validity_date: date | None = None) -> QuerySet:
        return self.get_queryset().filter_for_user(user, validity_date=validity_date)


class CommissionType(WBModel):
    id: Optional[int]
    name = models.CharField(max_length=256, verbose_name="Name")
    key = models.CharField(max_length=256, unique=True)

    def save(self, *args, **kwargs):
        if not self.key:
            self.key = self.name.lower()
        super().save(*args, **kwargs)

    def get_valid_commissions(
        self,
        terminal_account: "Account",
        compute_date: date,
        content_object: models.Model,
        terminal_accounts_total_holding: dict[int, Decimal],
    ) -> Generator[tuple["Commission", Decimal], None, None]:
        """
        Retrieve valid commissions for the given terminal account and parameters.

        This function filters and retrieves valid commissions for a specific terminal account
        based on the provided parameters. It iterates through the applicable commissions,
        validates them, and yields tuples of valid commissions along with their calculated percentages.

        Args:
            terminal_account (Account): The terminal account for which to retrieve valid commissions.
            compute_date (date): The date for which to compute the rebates.
            content_object (models.Model): The content object for which to compute rebates.
            terminal_accounts_total_holding (dict): The total holding for each terminal account of the root account.

        Yields:
            tuple: A tuple containing a valid Commission instance and its corresponding actual percentage.
        """

        available_percent = Decimal(1)
        # for hiearchical account structure, we prioritize the immidiate parent account (i.e. self)
        for commission in Commission.objects.filter(
            root_account=terminal_account.get_root(), commission_type=self, account__is_active=True
        ).order_by("order"):
            account_holding = sum(
                [
                    terminal_accounts_total_holding.get(account.id, Decimal("0"))
                    for account in commission.account.get_descendants(include_self=True)
                ]
            )
            commission.commission_account_holding = account_holding
            if commission.is_valid(compute_date, content_object, account_holding):
                actual_percent = (
                    available_percent * commission.validated_percent
                    if commission.validated_net_commission
                    else commission.validated_percent
                )
                actual_percent = max(min(available_percent, actual_percent), Decimal(0))
                available_percent -= actual_percent
                if actual_percent > 0:
                    yield commission, actual_percent

    def compute_rebates(
        self, root_account: "Account", verbose: bool = False, **iterator_kwargs
    ) -> Generator[tuple["Account", date, "Commission", models.Model, "Entry", Decimal, dict], None, None]:
        """
        Generate rebate information for terminal accounts based on given parameters.

        This method calculates and yields rebate information for terminal accounts based on the provided
        root_account, and iterator_kwargs.

        It iterates through active terminal accounts,
        computes rebates, and returns the results as a generator of tuples containing terminal account,
        compute date, commission, content object, recipient entry, and rebate amount.

        Args:
            root_account (Account): The root account for rebate calculations.
            verbose (bool): debugging option. Default to false.
            **iterator_kwargs: Additional keyword arguments for filtering.

        Yields:
            Generator[tuple]: A generator yielding rebate information tuples with components:
                - Terminal account (Account)
                - Compute date (date)
                - Commission (Commission)
                - Content object (models.Model)
                - Recipient entry (Entry)
                - Rebate amount (Decimal)

        """
        rebate_manager = AccountRebateManager(
            root_account, self.key
        )  # This will be loaded dynamically in a future iteration
        rebate_manager.initialize()
        # Iterate through active terminal accounts
        iterator = rebate_manager.get_iterator(**iterator_kwargs)
        if verbose:
            iterator = list(iterator)
            iterator = tqdm(iterator, total=len(iterator))

        for terminal_account, content_object, compute_date in iterator:
            terminal_accounts_total_holding = rebate_manager.get_terminal_accounts_total_holding(compute_date)
            terminal_account_holding_ratio = rebate_manager.get_terminal_account_holding_ratio(
                terminal_account,
                content_object,
                compute_date,
            )

            # Calculate rebates for the current terminal account
            if terminal_account_holding_ratio:
                for commission, actual_percent in self.get_valid_commissions(
                    terminal_account, compute_date, content_object, terminal_accounts_total_holding
                ):
                    # total commission pool for the given object that day
                    commission_pool = rebate_manager.get_commission_pool(content_object, compute_date)
                    commission_pool_sign = math.copysign(
                        1, commission_pool
                    )  # in case the pool is negative, the gain is a credit
                    commission_pool = abs(commission_pool)
                    account_gain = (
                        terminal_account_holding_ratio * commission_pool
                    )  # total account gain that account can get from the total commission pool for that day
                    if account_gain == 0:
                        continue

                    leftover_fees = account_gain

                    # Iterate through recipients and calculate rebates for each
                    for recipient, weighting in commission.get_recipients(
                        terminal_account, content_object, compute_date
                    ):
                        recipient_percent = actual_percent * weighting
                        recipient_fees = max(min(leftover_fees, account_gain * recipient_percent), Decimal(0))
                        leftover_fees -= recipient_fees

                        rebate_gain = Decimal(
                            math.copysign(recipient_fees, commission_pool_sign)
                        )  # If fees are negative, then we need to return the negative of the recipient fees

                        # Yield rebate information
                        yield (
                            terminal_account,
                            compute_date,
                            commission,
                            content_object,
                            recipient,
                            rebate_gain,
                            {
                                "terminal_account_holding_ratio": terminal_account_holding_ratio,
                                "commission_account_holding": getattr(commission, "commission_account_holding", None),
                                "commission_percent": actual_percent,
                                "commission_pool": commission_pool,
                                "recipient_percent": recipient_percent,
                            },
                        )

    def __str__(self) -> str:
        return self.name

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcommission:commissiontype"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcommission:commissiontyperepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{name}}"


class Commission(OrderedModel, WBModel):
    """Holds all the information on how fees are distributed"""

    id: Optional[int]
    account = models.ForeignKey[Account](
        to="wbcrm.Account",
        related_name="account_commissions",
        on_delete=models.CASCADE,
        verbose_name="Account",
    )
    root_account = models.ForeignKey[Account](
        to="wbcrm.Account",
        related_name="root_account_commissions",
        on_delete=models.CASCADE,
        verbose_name="Root Account",
    )  # we store the root account in order to allow reordering on root account as well as type

    crm_recipient = models.ForeignKey[Entry](
        to="directory.Entry",
        related_name="recipient_commissions",
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        verbose_name="Recipient",
    )
    portfolio_role_recipient = models.CharField(
        max_length=32, choices=PortfolioRole.RoleType.choices, null=True, blank=True, verbose_name="Recipient Role"
    )
    account_role_type_recipient = models.ForeignKey[AccountRoleType](
        to="wbcrm.AccountRoleType",
        related_name="recipient_commissions",
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        verbose_name="Account Role Type Recipient",
    )

    net_commission = models.BooleanField(default=True, verbose_name="Net Commission Rule")
    commission_type = models.ForeignKey[CommissionType](
        "wbcommission.CommissionType",
        on_delete=models.PROTECT,
        related_name="commissions",
        verbose_name="Commission Type",
    )
    is_hidden = models.BooleanField(
        default=True,
        verbose_name="Hidden Commission Rule",
        help_text="If False, this commission rule can be seen by anoyone that can access the related account. Otherwise, only an explicit role will grant access",
    )
    exclusion_rule_account_role_type = models.ForeignKey[AccountRoleType](  # TODO think if this is the best approach
        to="wbcrm.AccountRoleType",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name="Account Role to decide with exclusion rule applies",
    )
    objects = CommissionManager()
    order_with_respect_to = ("commission_type", "root_account")

    if TYPE_CHECKING:
        rules = RelatedManager["CommissionRule"]()

    class Meta(OrderedModel.Meta):
        verbose_name = "Commission"
        verbose_name_plural = "Commissions"
        permissions = (("administrate_commission", "Can administrate Commission"),)
        constraints = [
            UniqueConstraint(
                name="unique_crm_recipient_account",
                fields=["commission_type", "account", "crm_recipient"],
                condition=Q(crm_recipient__isnull=False),
            ),
            UniqueConstraint(
                name="unique_portfolio_role_recipient_account",
                fields=["commission_type", "account", "portfolio_role_recipient"],
                condition=Q(portfolio_role_recipient__isnull=False),
            ),
            UniqueConstraint(
                name="unique_account_role_type_recipient_account",
                fields=["commission_type", "account", "account_role_type_recipient"],
                condition=Q(account_role_type_recipient__isnull=False),
            ),
            models.CheckConstraint(
                condition=(
                    Q(
                        crm_recipient__isnull=False,
                        portfolio_role_recipient__isnull=True,
                        account_role_type_recipient__isnull=True,
                    )
                    | Q(
                        crm_recipient__isnull=True,
                        portfolio_role_recipient__isnull=False,
                        account_role_type_recipient__isnull=True,
                    )
                    | Q(
                        crm_recipient__isnull=True,
                        portfolio_role_recipient__isnull=True,
                        account_role_type_recipient__isnull=False,
                    )
                ),
                name="Only one recipient type set",
            ),
        ]

    def get_recipient_repr(self) -> str:
        if self.crm_recipient:
            return f"Profile {self.crm_recipient.computed_str}"
        elif self.portfolio_role_recipient:
            return f"Portfolio Role: {PortfolioRole.RoleType[self.portfolio_role_recipient].label}"
        else:
            return f"Account Role Type: {self.account_role_type_recipient.title}"

    def save(self, *args, **kwargs):
        self.root_account = self.account.get_root()
        super().save(*args, **kwargs)

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return f"Net: {self.net_commission} - {self.account.title} - {self.order} - {self.commission_type} - {self.get_recipient_repr()} (id: {self.id})"

    @property
    def validated_percent(self) -> Decimal:
        if not hasattr(self, "_validated_percent"):
            raise AssertionError("You must call `.is_valid()` before accessing `.validated_percent`.")
        return self._validated_percent

    @property
    def validated_net_commission(self) -> bool:
        if not hasattr(self, "_validated_net_commission"):
            raise AssertionError("You must call `.is_valid()` before accessing `.validated_net_commission`.")
        return self._validated_net_commission

    def is_valid(self, validity_date: date, product: models.Model, asset_under_management: Decimal) -> bool:
        """
        Check if the commission rule is valid for given parameters.

        This method determines if the commission rule is valid based on the provided parameters.
        It checks the validity date, the product, and the asset under management against the rules
        associated with the commission. If a valid rule is found, the method sets the validated
        percentage and net commission and handles any overriding exclusion rules.

        Args:
            validity_date (date): The date for which the validity is checked.
            product (models.Model): The product associated with the commission rule.
            asset_under_management (Decimal): The asset under management for the account.

        Returns:
            bool: True if the commission rule is valid, False otherwise.
        """
        try:
            valid_rule = self.rules.get(
                Q(timespan__startswith__lte=validity_date)
                & Q(timespan__endswith__gt=validity_date)
                & Q(assets_under_management_range__startswith__lte=asset_under_management)
                & (
                    Q(assets_under_management_range__endswith__gt=asset_under_management)
                    | Q(assets_under_management_range__endswith__isnull=True)
                )
                & (Q(percent__gt=0) | (Q(percent=0) & Q(consider_zero_percent_for_exclusion=True)))
            )
            self._validated_percent = valid_rule.percent
            self._validated_net_commission = self.net_commission
            with suppress(CommissionExclusionRule.DoesNotExist):
                overriding_rule = CommissionExclusionRule.objects.get(
                    timespan__startswith__lte=validity_date,
                    timespan__endswith__gt=validity_date,
                    product=product,
                    commission_type=self.commission_type,
                    account_role_type=self.exclusion_rule_account_role_type,
                )
                self._validated_percent = overriding_rule.overriding_percent
                self._validated_net_commission = overriding_rule.get_net_or_gross(self.net_commission)
            return True
        except CommissionRule.DoesNotExist:
            return False

    def get_recipients(
        self, account: "Account", product: models.Model, val_date: date
    ) -> Generator[tuple["Entry", Decimal], None, None]:
        """
        This function generates recipients and their respective weightings for distributing
        commissions based on the provided account, product, and validation date.

        Args:
            account (Account): The account for which recipients are being generated. Only used for account role type typed commission
            product (models.Model): The product model for which recipients are being generated. Only used for portfolio role typed commission
            val_date (date): The date for which the recipients are being generated. Used to determine the role validity

        Yields:
            tuple[Entry, Decimal]: A tuple containing a recipient entry and its associated weighting.

        Note:
            The generated recipients and weightings depend on different conditions, including CRM recipients,
            account role type recipients, and portfolio role recipients.

        Raises:
            AccountRoleValidity.DoesNotExist: If no validity data is available for any account role,
                                             the exception will be caught, and recipients will not be generated.
            KeyError: If no recipient data is available for any condition,
                      a KeyError will be caught, and no recipients will be generated.
        """
        if self.crm_recipient:
            yield self.crm_recipient, Decimal(1.0)
        elif self.account_role_type_recipient:
            account_roles = AccountRole.objects.annotate(
                is_valid=AccountRoleValidity.get_role_validity_subquery(val_date)
            ).filter(
                is_valid=True,
                account__in=account.get_ancestors(include_self=True).filter(level__gte=self.account.level),
                role_type=self.account_role_type_recipient,
            )
            total_weighting = Decimal(account_roles.aggregate(s=Sum("weighting"))["s"] or 0)
            for account_role in account_roles:
                yield (
                    account_role.entry,
                    Decimal(account_role.weighting) / total_weighting if total_weighting else Decimal(0),
                )
        else:
            role_recipients = PortfolioRole.objects.exclude(
                weighting=0
            ).filter(
                (Q(role_type=self.portfolio_role_recipient) & (Q(instrument=product) | Q(instrument__isnull=True)))
                & (Q(start__lte=val_date) | Q(start__isnull=True))
                & (
                    Q(end__gte=val_date) | Q(end__isnull=True)
                )  # This breaks the date range default upper range exclusion, will need to change if we ever move to a data range field
            )
            total_weighting = Decimal(role_recipients.aggregate(s=Sum("weighting"))["s"] or 0)
            for portfolio_role in role_recipients:
                yield (
                    portfolio_role.person.entry_ptr,
                    Decimal(Decimal(portfolio_role.weighting) / total_weighting) if total_weighting else Decimal(0),
                )

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcommission:commission"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcommission:commissionrepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"

    @classmethod
    def get_representation_label_key(cls) -> str:
        return "{{id}}"


class CommissionExclusionRule(models.Model):
    class NetOrGrossCommissionOverridingType(models.TextChoices):
        NET = "NET", "Net"
        GROSS = "GROSS", "Gross"
        DEFAULT = "DEFAULT", "Default"

    id: Optional[int]
    product = models.ForeignKey(
        to="wbportfolio.Product",
        on_delete=models.CASCADE,
        verbose_name="Product",
    )
    timespan = DateRangeField(verbose_name="Timespan")
    commission_type = models.ForeignKey(
        CommissionType,
        on_delete=models.PROTECT,
        related_name="commission_exclusion_rules",
        verbose_name="Commission Type",
    )
    account_role_type = models.ForeignKey(
        to="wbcrm.AccountRoleType",
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        verbose_name="Account Role Type Recipient",
    )
    overriding_percent = models.DecimalField(verbose_name="Overriding Percent", max_digits=4, decimal_places=3)
    overriding_net_or_gross_commission = models.CharField(
        default=NetOrGrossCommissionOverridingType.DEFAULT,
        choices=NetOrGrossCommissionOverridingType.choices,
        max_length=16,
        verbose_name="Overriding Net or Gross Commission Rule",
    )

    def save(self, *args, **kwargs):
        if not self.timespan:
            self.timespan = DateRange(date.min, date.max)  # type: ignore
        super().save(*args, **kwargs)

    def get_net_or_gross(self, net_commission: bool) -> bool:
        """
        Determine whether to use net or gross commission based on overriding settings.

        This function determines whether to use net or gross commission based on the
        provided `net_commission` parameter and the overriding settings.

        Args:
            net_commission (bool): The original commission type, True for net commission, False for gross commission.

        Returns:
            bool: The determined commission type after considering the overriding settings.
        """
        return {
            self.NetOrGrossCommissionOverridingType.DEFAULT.name: net_commission,
            self.NetOrGrossCommissionOverridingType.GROSS.name: False,
        }.get(self.overriding_net_or_gross_commission, True)

    class Meta:
        verbose_name = "Commission Exclusion Rule"
        verbose_name_plural = "Commissions Exclusion Rules"
        constraints = [
            ExclusionConstraint(
                name="exclude_overlapping_exclusion_rules",
                expressions=[
                    ("timespan", RangeOperators.OVERLAPS),
                    ("product", RangeOperators.EQUAL),
                    ("commission_type", RangeOperators.EQUAL),
                    ("account_role_type", RangeOperators.EQUAL),
                ],
            ),
        ]

    def __str__(self) -> str:
        return super().__str__()


class CommissionRule(ComplexToStringMixin, WBModel):
    id: Optional[int]
    commission = models.ForeignKey(
        Commission, on_delete=models.CASCADE, related_name="rules", verbose_name="Commission Line"
    )
    timespan = DateRangeField(verbose_name="Timespan")
    assets_under_management_range = DecimalRangeField(verbose_name="AUM Range")
    percent = models.DecimalField(verbose_name="Percent", default=Decimal(0), max_digits=4, decimal_places=3)
    consider_zero_percent_for_exclusion = models.BooleanField(
        default=False,
        verbose_name="Consider 0 percent for exclusion",
        help_text="If true, the commission rule with percent 0 (initially consider disabled), will be considered for exclusion rule matching",
    )

    def save(self, *args, **kwargs):
        if not self.timespan:
            self.timespan = DateRange(date.min, date.max)  # type: ignore
        super().save(*args, **kwargs)

    class Meta:
        constraints = [
            ExclusionConstraint(
                name="exclude_overlapping_rules",
                expressions=[
                    ("timespan", RangeOperators.OVERLAPS),
                    ("assets_under_management_range", RangeOperators.OVERLAPS),
                    ("commission", RangeOperators.EQUAL),
                ],
            ),
        ]

    def compute_str(self) -> str:
        return f"{self.commission} - {self.percent:.2%}: date range = [{self.timespan.lower} - {self.timespan.upper}[, aum range=[{self.assets_under_management_range.lower} - {self.assets_under_management_range.upper}["  # type: ignore

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcommission:commissionrule"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcommission:commissionrulerepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"


class CommissionRole(ComplexToStringMixin, WBModel):
    class Meta:
        verbose_name = "Commission Role"
        verbose_name_plural = "Commission Roles"
        constraints = [models.UniqueConstraint(fields=["person", "commission"], name="unique_commission_role")]

    def compute_str(self) -> str:
        return f"{self.person} -> {self.commission}"

    id: Optional[int]
    person = models.ForeignKey(
        "directory.Person", related_name="rebates_account_roles", on_delete=models.PROTECT, verbose_name="Person"
    )
    commission = models.ForeignKey(
        Commission, related_name="roles", on_delete=models.CASCADE, verbose_name="Commission Line"
    )

    @classmethod
    def get_endpoint_basename(cls) -> str:
        return "wbcommission:commissionrole"

    @classmethod
    def get_representation_endpoint(cls) -> str:
        return "wbcommission:commissionrolerepresentation-list"

    @classmethod
    def get_representation_value_key(cls) -> str:
        return "id"


@receiver(pre_merge, sender="directory.Person")
def pre_merge_person_commisionrole(sender: models.Model, merged_object, main_object, **kwargs):
    for role in CommissionRole.objects.filter(person=merged_object):
        if CommissionRole.objects.filter(person=main_object, commission=role.commission).exists():
            role.delete()
        else:
            role.person = main_object
            role.save()


@receiver(post_save, sender="wbcommission.Commission")
def post_commission_creation(sender, instance, created, **kwargs):
    # if a new commission line is created, we create a general rule
    if created:
        CommissionRule.objects.create(
            commission=instance,
            timespan=DateRange(date.min, date.max),  # type: ignore
            assets_under_management_range=NumericRange(0, None),  # type: ignore
        )


@receiver(pre_merge, sender="wbcrm.Account")
def handle_pre_merge_account_for_commission(
    sender: models.Model, merged_object: Account, main_object: Account, **kwargs
):
    """
    Loop over all commission lines assigned to the account to be merged and reassign them to the remaining account. Handle also the commission roles and rules.
    """
    for commission in merged_object.account_commissions.all():
        defaults = {
            "order": commission.order,
            "net_commission": commission.net_commission,
            "is_hidden": commission.is_hidden,
            "exclusion_rule_account_role_type": commission.exclusion_rule_account_role_type,
        }
        if commission.crm_recipient:
            new_commission, created = Commission.objects.get_or_create(
                crm_recipient=commission.crm_recipient,
                account=main_object,
                commission_type=commission.commission_type,
                defaults=defaults,
            )
        elif commission.portfolio_role_recipient:
            new_commission, created = Commission.objects.get_or_create(
                portfolio_role_recipient=commission.portfolio_role_recipient,
                account=main_object,
                commission_type=commission.commission_type,
                defaults=defaults,
            )
        else:
            new_commission, created = Commission.objects.get_or_create(
                account_role_type_recipient=commission.account_role_type_recipient,
                account=main_object,
                commission_type=commission.commission_type,
                defaults=defaults,
            )
        for role in commission.roles.all():
            if not new_commission.roles.filter(person=role.person).exists():
                role.commission = new_commission
                role.save()
        if created:
            new_commission.rules.all().delete()
        for rule in commission.rules.all():
            if not new_commission.rules.filter(
                timespan__overlap=rule.timespan,
                assets_under_management_range__overlap=rule.assets_under_management_range,
            ).exists():
                rule.commission = new_commission
                rule.save()
        commission.rebates.update(commission=new_commission)
        commission.delete()



from datetime import date
from io import BytesIO

import pandas as pd
from celery import shared_task
from wbcore.contrib.authentication.models import User
from wbcore.contrib.directory.models import Entry
from wbcore.workers import Queue
from wbcrm.models.accounts import Account
from wbportfolio.models.products import Product

from wbcommission.models import Commission, CommissionType, Rebate

from .utils import create_report_and_send


@shared_task(queue=Queue.DEFAULT.value)
def create_audit_report_and_send_as_task(user_id: int, recipient_id: int, start_date: date, end_date: date):
    user = User.objects.get(id=user_id)
    recipient = Entry.objects.get(id=recipient_id)
    create_report_and_send(user, recipient, start_date, end_date, create_report)


def create_report(user, customer, start_date, end_date):
    rebates = Rebate.objects.filter(recipient=customer, date__gte=start_date, date__lte=end_date).filter_for_user(user)
    df = pd.DataFrame(
        rebates.values(
            "date",
            "account",
            "product",
            "recipient",
            "commission_type",
            "commission",
            "value",
            "audit_log",
        )
    )
    buffer = BytesIO()
    if df.empty:
        raise ValueError("There is not rebate for this customer and given time period")
    df = pd.concat([df[df.columns.difference(["audit_log"])], pd.json_normalize(df["audit_log"])], axis=1).sort_values(
        by="date"
    )
    # we deserialize ids into human readable name
    df.commission_type = df.commission_type.map(dict(CommissionType.objects.values_list("id", "name")))
    df["product"] = df["product"].map(dict(Product.objects.values_list("id", "computed_str")))
    df.commission = df.commission.map(
        dict(map(lambda x: (x, str(Commission.objects.get(id=x))), df.commission.unique()))
    )
    df.recipient = df.recipient.map(dict(Entry.objects.values_list("id", "computed_str")))
    df.account = df.account.map(dict(Account.objects.values_list("id", "computed_str")))
    df.to_csv(buffer, index=False, mode="wb", encoding="UTF-8")
    return buffer, "audit_report_{}_{}_{}.csv".format(customer.computed_str, start_date, end_date), "application/csv"



# Generated by Django 5.2.9 on 2026-01-15 13:43

import django.db.models.deletion
from django.db import migrations, models

def migrate_root_account(apps, schema_editor):
    Commission = apps.get_model('wbcommission', 'Commission')
    Account = apps.get_model('wbcrm', 'Account')
    objs = []
    for c in Commission.objects.all():
        c.root_account = Account.objects.filter(tree_id=c.account.tree_id, level=0).first()
        objs.append(c)
    Commission.objects.bulk_update(objs, ["root_account"])

class Migration(migrations.Migration):

    dependencies = [
        ('wbcommission', '0011_commissionrole_unique_commission_role'),
        ('wbcrm', '0020_alter_product_unique_together_product_unique_product'),
    ]

    operations = [
        migrations.AddField(
            model_name='commission',
            name='root_account',
            field=models.ForeignKey(blank=True, null=True,on_delete=django.db.models.deletion.CASCADE, related_name='root_account_commissions', to='wbcrm.account', verbose_name='Root Account'),
            preserve_default=False,
        ),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL IMMEDIATE;"),
        migrations.RunPython(migrate_root_account),
        migrations.RunSQL(sql="SET CONSTRAINTS ALL DEFERRED;"),
        migrations.AlterField(
            model_name='commission',
            name='root_account',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE,
                                    related_name='root_account_commissions', to='wbcrm.account',
                                    verbose_name='Root Account'),
            preserve_default=False,
        ),
    ]



# Generated by Django 5.2.9 on 2025-12-09 08:50

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('wbcommission', '0008_alter_commission_options_alter_commission_order'),
    ]

    operations = [
        migrations.RenameIndex(
            model_name='rebate',
            new_name='wbcommissio_commiss_432402_idx',
            old_fields=('commission_type', 'date', 'recipient', 'product', 'account'),
        ),
    ]



from wbcommission.viewsets.endpoints.rebate import (
    RebatePandasViewEndpointConfig,
    RebateProductMarginalityEndpointConfig,
)
from .commissions import CommissionEndpointConfig, CommissionRuleCommissionEndpointConfig



from collections import defaultdict
from datetime import date, datetime
from decimal import Decimal
from functools import reduce
from typing import Iterable

from django.db.models import QuerySet, Sum
from wbaccounting.generators.base import AbstractBookingEntryGenerator
from wbaccounting.models import BookingEntry
from wbcore.contrib.directory.models import Entry
from wbportfolio.models import Product

from wbcommission.models import Rebate


class RebateGenerator(AbstractBookingEntryGenerator):
    TITLE = "Rebate Generation"

    @staticmethod
    def generate_booking_entries(
        from_date: date, to_date: date, counterparty: Entry, booking_date: date
    ) -> Iterable[BookingEntry]:
        rebates = Rebate.objects.filter(recipient=counterparty, date__gte=from_date, date__lte=to_date)
        vat = counterparty.entry_accounting_information.vat.get_rate(booking_date)
        for product_id in rebates.distinct("product_id").values_list("product_id", flat=True):
            product = Product.objects.get(id=product_id)

            for rebate_key, rebate_title in (("management", "Management"), ("performance", "Performance")):
                rebate_sum = (
                    rebates.filter(product_id=product_id, commission_type__key=rebate_key)
                    .aggregate(sum=Sum("value"))
                    .get("sum")
                    or 0.0
                )
                if rebate_sum > 0:
                    yield BookingEntry(
                        title=f"{product.name} {product.isin} {rebate_title} Fees",
                        booking_date=booking_date,
                        reference_date=to_date,
                        gross_value=Decimal(-1 * rebate_sum),
                        vat=vat,
                        currency=product.currency,
                        counterparty=counterparty,
                        parameters={
                            "from_date": from_date.strftime("%d.%m.%Y"),
                            "to_date": to_date.strftime("%d.%m.%Y"),
                        },
                        backlinks={
                            "comm-rebates": {
                                "title": "Rebates",
                                "reverse": "wbcommission:rebatetable-list",
                                "parameters": {
                                    "date": f"{from_date.strftime('%Y-%m-%d')},{to_date.strftime('%Y-%m-%d')}",
                                    "product": product_id,
                                    "recipient": counterparty.id,
                                    "group_by": "ACCOUNT",
                                },
                            },
                        },
                    )

    @staticmethod
    def _compare(d1: str, d2: str) -> str:
        d1_lb, d1_ub = d1.split(",")
        d2_lb, d2_ub = d2.split(",")

        lb = min(datetime.strptime(d1_lb, "%Y-%m-%d"), datetime.strptime(d2_lb, "%Y-%m-%d"))
        ub = max(datetime.strptime(d1_ub, "%Y-%m-%d"), datetime.strptime(d2_ub, "%Y-%m-%d"))

        return f"{lb.strftime('%Y-%m-%d')},{ub.strftime('%Y-%m-%d')}"

    @staticmethod
    def _merge_key_into_dict(d: dict, key: str, value: dict) -> dict:
        d[key]["title"] = value["title"]
        d[key]["reverse"] = value["reverse"]
        if "recipient" in value["parameters"]:
            d[key]["parameters"]["recipient"] = value["parameters"]["recipient"]
            d[key]["parameters"]["date"] = RebateGenerator._compare(
                d[key]["parameters"].get("date", "9999-12-31,1111-01-01"), value["parameters"]["date"]
            )
        return d

    @staticmethod
    def _iter_backlinks(booking_entries: QuerySet) -> Iterable:
        for backlink in booking_entries:
            yield from backlink.items()

    @staticmethod
    def merge_backlinks(booking_entries: QuerySet[BookingEntry]) -> dict:
        return reduce(
            lambda d, v: RebateGenerator._merge_key_into_dict(d, *v),
            RebateGenerator._iter_backlinks(
                booking_entries.filter(backlinks__isnull=False).values_list("backlinks", flat=True)
            ),
            defaultdict(lambda: dict(parameters=dict())),
        )



from django.dispatch import receiver
from wbcore.contrib.directory.viewsets import (
    CompanyModelViewSet,
    EntryModelViewSet,
    PersonModelViewSet,
)
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.display.instance_display.shortcuts import (
    create_simple_display,
)
from wbcore.signals.instance_buttons import add_instance_button
from wbcore.utils.serializers import StartEndDateSerializer


@receiver(add_instance_button, sender=PersonModelViewSet)
@receiver(add_instance_button, sender=EntryModelViewSet)
@receiver(add_instance_button, sender=CompanyModelViewSet)
def crm_adding_instance_buttons(sender, many, *args, **kwargs):
    if many:
        return
    return bt.DropDownButton(
        label="Commission",
        icon=WBIcon.UNFOLD.icon,
        buttons=(
            bt.WidgetButton(key="rebates", label="Rebates", icon=WBIcon.DEAL_MONEY.icon),
            bt.ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcommission:rebates",),
                action_label="Send Customer Report",
                key="generate_customer_commission_report",
                description_fields="<p>Generate Commission Report</p>",
                serializer=StartEndDateSerializer,
                icon=WBIcon.CHART_SWITCHES.icon,
                title="Generate Commission reports",
                label="Generate Commission reports",
                instance_display=create_simple_display([["start"], ["end"]]),
            ),
            bt.ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbcommission:rebates",),
                action_label="Send Audit Report",
                key="generate_audit_commission_report",
                description_fields="<p>Generate Audit Commission Report</p>",
                serializer=StartEndDateSerializer,
                icon=WBIcon.NOTEBOOK.icon,
                title="Generate Audit Commission reports",
                label="Generate Audit Commission reports",
                instance_display=create_simple_display([["start"], ["end"]]),
            ),
        ),
    )



# Generated by Django 5.2.9 on 2025-12-16 15:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('directory', '0014_alter_entry_relationship_managers_and_more'),
        ('wbcommission', '0009_rename_rebate_commission_type_date_recipient_product_account_wbcommissio_commiss_432402_idx'),
        ('wbcrm', '0019_alter_activity_companies_alter_activity_participants_and_more'),
        ('wbportfolio', '0095_portfolio_keep_only_essential_positions'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='rebate',
            unique_together=set(),
        ),
        migrations.AddConstraint(
            model_name='rebate',
            constraint=models.UniqueConstraint(fields=('date', 'recipient', 'account', 'product', 'commission_type'), name='unique_rebate'),
        ),
    ]



import pytest
from django.contrib.auth.models import Permission
from faker import Faker
from rest_framework.reverse import reverse
from rest_framework.test import APIClient
from wbcore.contrib.authentication.factories import UserFactory
from wbcore.contrib.authentication.models import User
from wbcore.contrib.permission.tests.test_backend import create_internal_user

from wbcommission.permissions import IsCommissionAdmin

fake = Faker()


@pytest.mark.django_db
class TestPermissionClass:
    @pytest.fixture()
    def admin_user(self):
        user = UserFactory.create()
        user.user_permissions.add(
            Permission.objects.get(content_type__app_label="wbcommission", codename="administrate_commission")
        )
        return User.objects.get(id=user.id)

    def test_normal_user_permission(self, rf, user):
        rf.user = user
        assert not IsCommissionAdmin().has_permission(rf, None)

    def test_admin_permission(self, rf, admin_user):
        rf.user = admin_user
        assert IsCommissionAdmin().has_permission(rf, None)

    def test_recompute_only_for_admin(self, user, admin_user):
        client = APIClient()
        url = reverse("wbcommission:rebate-recompute")

        client.force_authenticate(user=user)
        response = client.patch(url, data={"start_date": fake.date_object()})
        assert response.status_code == 403

        client.force_authenticate(user=admin_user)
        response = client.patch(url, data={"start_date": fake.date_object()})
        assert response.status_code == 200

    def test_auditreport_only_for_admin(self, user, admin_user, company):
        client = APIClient()
        url = reverse("wbcommission:rebate-auditreport")

        client.force_authenticate(user=user)
        response = client.patch(url + f"?recipient_id={company.id}")
        assert response.status_code == 403

        client.force_authenticate(user=admin_user)
        response = client.patch(url + f"?recipient_id={company.id}")
        assert response.status_code == 200

    def test_customerreport_only_for_internaluser(self, user, admin_user, company):
        client = APIClient()
        url = reverse("wbcommission:rebate-customerreport")

        client.force_authenticate(user=user)
        response = client.patch(url + f"?recipient_id={company.id}")
        assert response.status_code == 403

        client.force_authenticate(user=create_internal_user())
        response = client.patch(url + f"?recipient_id={company.id}")
        assert response.status_code == 200



from wbcore.menus import ItemPermission, MenuItem

COMMISSION_MENUITEM = MenuItem(
    label="Commissions",
    endpoint="wbcommission:commission-list",
    permission=ItemPermission(permissions=["wbcommission.view_commission"]),
)



from datetime import date

from django.db.models import Exists, OuterRef, Q
from wbcore import filters as wb_filters
from wbportfolio.models import Product

from wbcommission.models import Commission, CommissionRule


def filter_aum(queryset, value):
    return queryset.filter(
        Q(assets_under_management_range__startswith__lte=value)
        & (
            Q(assets_under_management_range__endswith__gt=value)
            | Q(assets_under_management_range__endswith__isnull=True)
        )
        & (Q(percent__gt=0) | Q(consider_zero_percent_for_exclusion=True))
    )


def filter_timestamp(queryset, value):
    return queryset.filter(
        Q(timespan__startswith__lte=value)
        & Q(timespan__endswith__gt=value)
        & (Q(percent__gt=0) | Q(consider_zero_percent_for_exclusion=True))
    )


class CommissionRuleCommissionFilterSet(wb_filters.FilterSet):
    only_valid_at_date = wb_filters.DateFilter(
        required=False,
        initial=lambda f, r, v: date.today(),
        method="filter_only_valid_at_date",
        label="Only Valid at date",
        help_text="Set to show only rules that are valid for the given date",
    )

    def filter_only_valid_at_date(self, queryset, name, value):
        if value:
            return filter_timestamp(queryset, value)
        return queryset

    only_valid_for_aum = wb_filters.NumberFilter(
        required=False,
        method="filter_only_valid_for_aum",
        label="Only Valid for AUM",
        help_text="Set to show only rules that are valid for the given aum",
    )

    def filter_only_valid_for_aum(self, queryset, name, value):
        if value:
            return filter_aum(queryset, value)
        return queryset

    class Meta:
        model = CommissionRule
        fields = {
            "commission": ["exact"],
        }


class CommissionAccountFilterSet(wb_filters.FilterSet):
    only_valid_at_date = wb_filters.DateFilter(
        required=False,
        initial=lambda f, r, v: date.today(),
        method="filter_only_valid_at_date",
        label="Only Valid at date",
        help_text="Set to show only rules that are valid for the given date",
    )

    def filter_only_valid_at_date(self, queryset, name, value):
        if value:
            return queryset.annotate(
                has_valid_rule=Exists(
                    filter_timestamp(CommissionRule.objects.filter(commission=OuterRef("pk")), value)
                )
            ).filter(has_valid_rule=True)

        return queryset

    only_valid_for_aum = wb_filters.NumberFilter(
        required=False,
        method="filter_only_valid_for_aum",
        label="Only Valid for AUM",
        help_text="Set to show only rules that are valid for the given aum",
    )

    def filter_only_valid_for_aum(self, queryset, name, value):
        if value:
            return queryset.annotate(
                has_valid_rule=Exists(filter_aum(CommissionRule.objects.filter(commission=OuterRef("pk")), value))
            ).filter(has_valid_rule=True)

        return queryset

    simulate_for_product = wb_filters.ModelChoiceFilter(
        label="Simulate for Product",
        queryset=Product.objects.all(),
        endpoint=Product.get_representation_endpoint(),
        value_key=Product.get_representation_value_key(),
        label_key=Product.get_representation_label_key(),
        method="fake_filter",
        depends_on=[{"field": "only_valid_at_date", "options": {}}],
        filter_params=lambda request, view: {"with_claim_for_account": view.account.id},
        help_text='Choose a product to compute the actual percent paid for the bellow recipient (need the "Valide at date" filter)',
    )

    class Meta:
        model = Commission
        fields = {}



from rest_framework.reverse import reverse
from wbcore.metadata.configs.endpoints import EndpointViewConfig


class CommissionEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse(
            "wbcommission:type-account-commission-list",
            args=[self.view.kwargs["commission_type_id"], self.view.kwargs["account_id"]],
            request=self.request,
        )


class CommissionRuleCommissionEndpointConfig(EndpointViewConfig):
    def get_endpoint(self, **kwargs):
        return reverse("wbcommission:commission-rule-list", args=[self.view.commission.id], request=self.request)



from collections import defaultdict
from datetime import date
from decimal import Decimal

import numpy as np
import pandas as pd
from django.contrib.messages import warning
from django.db.models import Exists, ExpressionWrapper, F, FloatField, OuterRef, Sum
from django.db.models.functions import Coalesce
from django.shortcuts import get_object_or_404
from django.utils.dateparse import parse_date
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response
from wbcore import serializers as wb_serializers
from wbcore import viewsets
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.directory.models import Entry
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.contrib.pandas import fields as pf
from wbcore.contrib.permission.permissions import IsInternalUser
from wbcore.serializers import decorator
from wbcore.utils.date import get_date_interval_from_request
from wbcore.utils.strings import format_number
from wbcrm.models.accounts import Account
from wbfdm.models import Classification, ClassificationGroup, InstrumentPrice
from wbfdm.preferences import get_default_classification_group
from wbportfolio.models import Product

from wbcommission.filters import (
    CustomerRebateGroupByFilter,
    RebateDateFilter,
    RebateGroupByFilter,
    RebateMarginalityFilter,
)
from wbcommission.models import CommissionType, Rebate
from wbcommission.models.rebate import RebateGroupbyChoice, manage_rebate_as_task
from wbcommission.reports.audit_report import create_audit_report_and_send_as_task
from wbcommission.reports.customer_report import create_customer_report_and_send_as_task
from wbcommission.serializers import RebateModelSerializer
from wbcommission.viewsets.buttons.rebate import RebateTableButtonConfig
from wbcommission.viewsets.display.rebate import (
    RebatePandasViewDisplayConfig,
    RebateProductMarginalityDisplayConfig,
)
from wbcommission.viewsets.endpoints.rebate import (
    RebatePandasViewEndpointConfig,
    RebateProductMarginalityEndpointConfig,
)
from wbcommission.viewsets.titles.rebate import (
    RebatePandasViewTitleConfig,
    RebateProductMarginalityTitleConfig,
)

from ..analytics.marginality import MarginalityCalculator
from ..permissions import IsCommissionAdmin
from .mixins import RebatePermissionMixin


class RebateModelViewSet(RebatePermissionMixin, viewsets.ModelViewSet):
    serializer_class = RebateModelSerializer
    queryset = Rebate.objects.all()
    filterset_class = RebateDateFilter

    endpoint_config_class = RebatePandasViewEndpointConfig

    def get_aggregates(self, queryset, paginated_queryset):
        return {"value_usd": {"Σ": format_number(queryset.aggregate(s=Sum(F("value_usd")))["s"] or 0)}}

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(
                fx_rate=Coalesce(CurrencyFXRates.get_fx_rates_subquery("date"), Decimal(1)),
                value_usd=ExpressionWrapper(F("value") * F("fx_rate"), output_field=FloatField()),
            )
        )

    @action(detail=False, methods=["PATCH"], permission_classes=[IsCommissionAdmin])
    def recompute(self, request, pk=None):
        if "start_date" in request.POST:
            prune_existing = request.POST.get("prune_existing", "false") == "true"
            start_date = parse_date(request.POST["start_date"])
            if only_account_ids_repr := request.POST.get("only_accounts", None):
                only_account_ids = only_account_ids_repr.split(",")
                accounts = Account.objects.filter(id__in=only_account_ids, level=0)
            else:
                accounts = Account.objects.filter(level=0)
            for account in accounts:
                manage_rebate_as_task.delay(
                    account.id, start_date=start_date, prune_existing=prune_existing, user=request.user
                )
            return Response({"status": f"{accounts.count()} are being recomputed"}, status=status.HTTP_200_OK)
        return Response({}, status=status.HTTP_400_BAD_REQUEST)

    @action(detail=False, methods=["GET", "PATCH"], permission_classes=[IsInternalUser])
    def customerreport(self, request, pk=None):
        start, end = get_date_interval_from_request(request, request_type="POST")
        user = request.user
        recipient = get_object_or_404(Entry, pk=request.GET.get("recipient_id", None))
        create_customer_report_and_send_as_task.delay(user.id, recipient.id, start, end)
        return Response({"__notification": {"title": "Report send."}}, status=status.HTTP_200_OK)

    @action(detail=False, methods=["GET", "PATCH"], permission_classes=[IsCommissionAdmin])
    def auditreport(self, request, pk=None):
        start, end = get_date_interval_from_request(request, request_type="POST")
        user = request.user
        recipient = get_object_or_404(Entry, pk=request.GET.get("recipient_id", None))
        create_audit_report_and_send_as_task.delay(user.id, recipient.id, start, end)
        return Response({"__notification": {"title": "Report send."}}, status=status.HTTP_200_OK)


class RebatePandasView(RebatePermissionMixin, ExportPandasAPIViewSet):
    queryset = Rebate.objects.all()

    filterset_class = RebateGroupByFilter

    search_fields = ["title"]

    display_config_class = RebatePandasViewDisplayConfig
    title_config_class = RebatePandasViewTitleConfig
    endpoint_config_class = RebatePandasViewEndpointConfig
    button_config_class = RebateTableButtonConfig

    def get_ordering_fields(self):
        return ["rebate_total", *map(lambda x: f"rebate_{x}", self.rebate_types.keys())]

    def get_pandas_fields(self, request):
        fields = [pf.PKField(key="id", label="ID"), pf.CharField(key="title", label="Title")]
        for key, label in self.rebate_types.items():
            fields.append(
                pf.FloatField(
                    key="rebate_" + key,
                    label=label,
                    precision=2,
                    decorators=[decorator(decorator_type="text", position="left", value="$")],
                )
            )
        fields.append(
            pf.FloatField(
                key="rebate_total",
                label="Total Rebate",
                precision=2,
                decorators=[decorator(decorator_type="text", position="left", value="$")],
            )
        )
        return pf.PandasFields(fields=fields)

    @cached_property
    def groupby_classification_group(self) -> ClassificationGroup:
        try:
            return ClassificationGroup.objects.get(id=self.request.GET["groupby_classification_group"])
        except (ValueError, KeyError):
            return get_default_classification_group()

    @cached_property
    def rebate_types(self):
        return dict(CommissionType.objects.values_list("key", "name"))

    def get_queryset(self):
        qs = (
            super()
            .get_queryset()
            .annotate(
                fx_rate=CurrencyFXRates.get_fx_rates_subquery("date", lookup_expr="exact"),
                value_usd=F("value") * F("fx_rate"),
            )
        )
        qs = Account.annotate_root_account_info(qs)
        qs = self.groupby_classification_group.annotate_queryset(qs, 0, "product")
        return (
            qs.select_related("product")
            .prefetch_related("product__currency")
            .prefetch_related("product__parent")
            .select_related("recipient")
        )

    def get_dataframe(self, request, queryset, **kwargs):
        groupby = self.request.GET.get("group_by", "PRODUCT")
        groupby_map = RebateGroupbyChoice.map[groupby]
        pivot = groupby_map["pk"]
        pivot_label = groupby_map["title_key"]
        if pivot == "classification_id":
            df = pd.DataFrame(
                queryset.values_list(
                    "value_usd",
                    "commission_type__key",
                    "classifications",
                ),
                columns=["value_usd", "commission_type__key", "classifications"],
            )
            df = (
                df.explode("classifications")
                .rename(columns={"classifications": "classification_id"})
                .replace([np.inf, -np.inf, np.nan], None)
            )
            df["classification_title"] = df.classification_id.map(
                dict(Classification.objects.filter(id__in=df.classification_id.unique()).values_list("id", "name")),
                na_action="ignore",
            )
        else:
            df = pd.DataFrame(
                queryset.values_list(
                    "value_usd",
                    "commission_type__key",
                    pivot,
                    pivot_label,
                ),
                columns=["value_usd", "commission_type__key", pivot, pivot_label],
            )
        df = df.rename(columns={pivot: "id", pivot_label: "title"})
        df = (
            pd.pivot_table(
                df,
                index=["id", "title"],
                columns="commission_type__key",
                values="value_usd",
                aggfunc="sum",
            )
            .astype(float)
            .fillna(0)
        )
        df["total"] = df.sum(axis=1)
        # Ensure all type are always present as a columns
        for commission_type in self.rebate_types.keys():
            if commission_type not in df.columns:
                df[commission_type] = 0
        df = df.add_prefix("rebate_").reset_index()
        df.id = df.id.fillna(0)
        df.title = df.title.fillna("None")
        df[df.columns.difference(["id", "title"])] = df[df.columns.difference(["id", "title"])].astype("float")
        return df

    def get_aggregates(self, request, df):
        if df.empty:
            return {}
        return {
            **{
                "rebate_" + commission_type.key: {
                    "Σ": format_number(df["rebate_" + commission_type.key].sum()),
                }
                for commission_type in CommissionType.objects.all()
            },
            "rebate_total": {
                "Σ": format_number(df["rebate_total"].sum()),
            },
        }

    def get_filterset_class(self, request):
        profile = request.user.profile
        if profile.is_internal or request.user.is_superuser:
            return RebateGroupByFilter
        return CustomerRebateGroupByFilter


class RebateProductMarginalityViewSet(ExportPandasAPIViewSet):
    IDENTIFIER = "wbcommission:product-marginality"

    filterset_class = RebateMarginalityFilter

    pandas_fields = pf.PandasFields(
        fields=[
            pf.PKField(key="id", label="ID"),
            pf.CharField(key="title", label="title"),
            pf.CharField(key="currency_symbol", label="Currency"),
            pf.FloatField(key="base_management_fees_percent", label="Base Management Fees", percent=True),
            pf.FloatField(
                key="management_fees",
                label="management_fees",
                decorators=[
                    wb_serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}")
                ],
            ),
            pf.FloatField(
                key="management_rebates",
                label="management_rebates",
                decorators=[
                    wb_serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}")
                ],
            ),
            pf.FloatField(
                key="management_marginality",
                label="management_marginality",
                decorators=[
                    wb_serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}")
                ],
            ),
            pf.FloatField(key="base_performance_fees_percent", label="Base Performance Fees", percent=True),
            pf.FloatField(
                key="performance_fees",
                label="performance_fees",
                decorators=[
                    wb_serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}")
                ],
            ),
            pf.FloatField(
                key="performance_rebates",
                label="performance_rebates",
                decorators=[
                    wb_serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}")
                ],
            ),
            pf.FloatField(
                key="performance_marginality",
                label="performance_marginality",
                decorators=[
                    wb_serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}")
                ],
            ),
            pf.FloatField(
                key="total_fees",
                label="total_fees",
                decorators=[
                    wb_serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}")
                ],
            ),
            pf.FloatField(
                key="total_rebates",
                label="total_rebates",
                decorators=[
                    wb_serializers.decorator(decorator_type="text", position="left", value="{{currency_symbol}}"),
                ],
            ),
            pf.FloatField(key="total_marginality_percent", label="total_marginality_percent", percent=True),
            pf.FloatField(
                key="total_fees_usd",
                label="total_fees_usd",
                decorators=[wb_serializers.decorator(decorator_type="text", position="left", value="$")],
            ),
            pf.FloatField(
                key="total_rebates_usd",
                label="total_rebates_usd",
                decorators=[wb_serializers.decorator(decorator_type="text", position="left", value="$")],
            ),
            pf.FloatField(
                key="total_marginality_usd",
                label="total_marginality_usd",
                decorators=[wb_serializers.decorator(decorator_type="text", position="left", value="$")],
            ),
            pf.FloatField(key="net_management_marginality", label="net_management_marginality", percent=True),
            pf.FloatField(key="net_performance_marginality", label="net_performance_marginality", percent=True),
        ]
    )
    queryset = Product.objects.all()
    ordering_fields = [
        "title",
        "base_management_fees_percent",
        "base_performance_fees_percent",
        "management_fees",
        "management_rebates",
        "management_marginality",
        "performance_fees",
        "performance_rebates",
        "performance_marginality",
        "total_fees",
        "total_rebates",
        "total_marginality_percent",
        "total_fees_usd",
        "total_rebates_usd",
        "total_marginality_usd",
        "net_management_marginality",
    ]
    search_fields = ["title", "bank_title"]
    ordering = ["title"]

    display_config_class = RebateProductMarginalityDisplayConfig
    title_config_class = RebateProductMarginalityTitleConfig
    endpoint_config_class = RebateProductMarginalityEndpointConfig

    def get_aggregates(self, request, df):
        aggregates = {}
        if not df.empty:
            total_fees = df.total_fees.sum()
            total_rebates = df.total_rebates.sum()
            total_marginality_percent = 1 - total_rebates / total_fees if total_fees != 0 else 0

            aggregates = {
                "total_marginality_percent": {"Σ": format_number(total_marginality_percent)},
                "total_fees_usd": {"Σ": format_number(df.total_fees_usd.sum())},
                "total_rebates_usd": {"Σ": format_number(df.total_rebates_usd.sum())},
                "total_marginality_usd": {"Σ": format_number(df.total_marginality_usd.sum())},
            }
            if aggregated_net_management_marginality := getattr(self, "aggregated_net_management_marginality", None):
                aggregates["net_management_marginality"] = {
                    "Σ": format_number(aggregated_net_management_marginality, decimal=4)
                }
            if aggregated_net_performance_marginality := getattr(self, "aggregated_net_performance_marginality", None):
                aggregates["net_performance_marginality"] = {
                    "Σ": format_number(aggregated_net_performance_marginality, decimal=4)
                }
            currency_aggregates = defaultdict(dict)
            for currency_symbol in df["currency_symbol"].unique():
                for field in [
                    "management_fees",
                    "management_rebates",
                    "performance_fees",
                    "performance_rebates",
                    "total_fees",
                    "total_rebates",
                    "management_marginality",
                    "performance_marginality",
                ]:
                    currency_aggregates[field][currency_symbol] = format_number(
                        df.loc[df["currency_symbol"] == currency_symbol, field].sum()
                    )
            aggregates.update(currency_aggregates)
        return aggregates

    @cached_property
    def start(self) -> date | None:
        return get_date_interval_from_request(self.request, exclude_weekend=False)[0]

    @cached_property
    def end(self) -> date | None:
        return get_date_interval_from_request(self.request, exclude_weekend=False)[1]

    def get_queryset(self):
        if self.request.user.has_perm("wbcommission.administrate_commission"):
            has_aum_subquery = InstrumentPrice.objects.filter(instrument=OuterRef("pk"), outstanding_shares__gt=0)
            if self.start and self.end:
                has_aum_subquery = has_aum_subquery.filter(
                    date__gte=self.start,
                    date__lte=self.end,
                )
            return super().get_queryset().annotate(has_aum=Exists(has_aum_subquery))
        return Product.objects.none()

    def get_dataframe(self, request, queryset, **kwargs):
        df = pd.DataFrame()
        if queryset.exists() and self.start and self.end:
            if self.start.weekday() in (5, 6) or self.end.weekday() in (5, 6):
                warning(
                    request,
                    _(
                        "The selected date range includes a Saturday or Sunday. Please note that fees and rebates are normalized over the weekend, as fees continue to accumulate during this period."
                    ),
                )
            marginality_calculator = MarginalityCalculator(queryset, self.start, self.end)
            df_products = (
                pd.DataFrame(
                    queryset.values(
                        "id", "computed_str", "current_management_fees", "current_performance_fees", "currency__symbol"
                    )
                )
                .set_index("id")
                .rename(
                    columns={
                        "currency__symbol": "currency_symbol",
                        "computed_str": "title",
                        "current_management_fees": "base_management_fees_percent",
                        "current_performance_fees": "base_performance_fees_percent",
                    }
                )
            )
            # Concat dataframe and add extra statistic columns
            df = pd.concat(
                [
                    df_products["title"],
                    df_products["base_management_fees_percent"],
                    df_products["base_performance_fees_percent"],
                    df_products["currency_symbol"],
                    marginality_calculator.management_fees,
                    marginality_calculator.performance_fees,
                    marginality_calculator.total_fees,
                    marginality_calculator.total_rebates,
                    marginality_calculator.total_marginality_percent,
                    marginality_calculator.management_rebates,
                    marginality_calculator.performance_rebates,
                    marginality_calculator.management_marginality,
                    marginality_calculator.performance_marginality,
                    marginality_calculator.management_marginality_percent,
                    marginality_calculator.total_fees_usd,
                    marginality_calculator.total_rebates_usd,
                    marginality_calculator.total_marginality_usd,
                ],
                axis=1,
            )
            df["net_management_marginality"] = marginality_calculator.get_net_marginality("management")
            df["net_performance_marginality"] = marginality_calculator.get_net_marginality("performance")
            self.aggregated_net_management_marginality = marginality_calculator.get_aggregated_net_marginality(
                "management"
            )
            self.aggregated_net_performance_marginality = marginality_calculator.get_aggregated_net_marginality(
                "performance"
            )
            # Sanitize
            df = df.replace([np.inf, -np.inf, np.nan], 0).reset_index()
            df = df[(df["total_fees"] != 0) | (df["total_rebates"] != 0)]
            return df
        return df



# Generated by Django 5.2.9 on 2026-01-14 09:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('directory', '0015_alter_emailcontact_address_and_more'),
        ('wbcommission', '0010_alter_rebate_unique_together_rebate_unique_rebate'),
    ]

    operations = [
        migrations.AddConstraint(
            model_name='commissionrole',
            constraint=models.UniqueConstraint(fields=('person', 'commission'), name='unique_commission_role'),
        ),
    ]



from .buttons import *  # noqa
from .rebate import (
    RebateModelViewSet,
    RebatePandasView,
    RebateProductMarginalityViewSet,
)
from .commissions import (
    CommissionTypeRepresentationModelViewSet,
    CommissionTypeModelViewSet,
    CommissionModelViewSet,
    CommissionRuleCommissionModelViewSet,
    CommissionRepresentationModelViewSet
)
